/*
 * SPDX-FileCopyrightText: 2016-2016 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 */

#include "userinterface.h"

namespace fcitx {

UserInterface::~UserInterface() {}

VirtualKeyboardUserInterface::~VirtualKeyboardUserInterface() {}

} // namespace fcitx
